/*
 * (C) 1997 Florian Hars
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  This software is provided "as is" without express or
 * implied warranty.
 */

#include "pgmrib.h"

RtPoint *subv(RtPoint *ret, RtPoint *arg1, RtPoint *arg2) {
  int i;

  for (i=0; i<3; i++) {
    (*ret)[i]=(*arg1)[i] - (*arg2)[i];
  }
  return ret;
}

RtPoint *unitnorm(RtPoint *ret, RtPoint *arg1, RtPoint *arg2) {
  int i;
  RtFloat length;

    (*ret)[0] = (*arg1)[1] * (*arg2)[2] - (*arg1)[2] * (*arg2)[1];
    (*ret)[1] = (*arg1)[2] * (*arg2)[0] - (*arg1)[0] * (*arg2)[2];
    (*ret)[2] = (*arg1)[0] * (*arg2)[1] - (*arg1)[1] * (*arg2)[0];
    length =0 ;
    for (i=0; i<3; i++) {
      length += (*ret)[i] * (*ret)[i];
    }
    length = sqrt(length);
    if (length != 0) {
      for (i=0; i<3; i++) {
	(*ret)[i] = (*ret)[i]/length;
      }
    } else {
      fprintf(stderr, "Undefined normal, using (0, 0, 0)\n");
    }
    return ret;
}

extern int parsepgm(gray **pgmmesh, int cols, int rows, gray maxval) {
  RtFloat *mesh;
  RtPoint *points, *normals, ret1, ret2;
  int i,u,v;

  if ((mesh=(RtFloat *) malloc(cols*rows*sizeof(RtFloat))) == NULL) {
    fprintf(stderr, "Could not allocate mesh: Malloc failed\n");
    return(-1);
  }
  if ((points=(RtPoint *) malloc(cols*rows*sizeof(RtPoint))) == NULL) {
    fprintf(stderr, "Malloc failed for points\n");
    free(mesh);
    return(-1);
  }
  if ((normals=(RtPoint *) malloc((cols*rows)*sizeof(RtPoint))) == NULL) {
    fprintf(stderr, "Malloc failed for normals\n");
    free(points);
    free(mesh);
    return(-1);
  }

#define idx(a,b)  (a + (b) * cols)
 
  for (v=0; v < rows; v++) {
    for (u=0; u < cols; u++) {
      mesh[idx(u,v)] =  (RtFloat) pgmmesh[v][u]/maxval;
      points[idx(u,v)][0] = (RtFloat) u/(cols-1);
      points[idx(u,v)][1] = (RtFloat) v/(rows-1);
      points[idx(u,v)][2] = mesh[idx(u,v)];
    }
  }
  (void) unitnorm(&normals[idx(0,0)], subv(&ret1, &points[idx(1,1)], &points[idx(0,0)]),
			   subv(&ret2, &points[idx(0,1)], &points[idx(1,0)]));
  (void) unitnorm(&normals[idx(cols-1,0)], subv(&ret1, &points[idx(cols-2,1)],
						&points[idx(cols-1,0)]),
				       subv(&ret2, &points[idx(cols-2,0)],
						&points[idx(cols-1,1)]));
  (void) unitnorm(&normals[idx(0,rows-1)], subv(&ret1, &points[idx(1,rows-2)],
					     &points[idx(0,rows-1)]),
				    subv(&ret2, &points[idx(1,rows-1)],
					     &points[idx(0,rows-2)]));
  (void) unitnorm(&normals[idx(cols-1,rows-1)], subv(&ret1, &points[idx(cols-2,rows-2)], 
					   &points[idx(cols-1,rows-1)]),
				  subv(&ret2, &points[idx(cols-1,rows-2)],
					   &points[idx(cols-2,rows-1)]));
  for (u=1; u<cols-1; u++){
    (void) unitnorm(&normals[idx(u,0)], subv(&ret1, &points[idx(u+1,1)], &points[idx(u,0)]),
			  subv(&ret2, &points[idx(u-1,1)], &points[idx(u,0)]));
    (void) unitnorm(&normals[idx(u,rows-1)], subv(&ret1, &points[idx(u-1,rows-2)],
							   &points[idx(u,rows-1)]),
						  subv(&ret2, &points[idx(u+1,rows-2)],
							   &points[idx(u,rows-1)]));
  }
  for (v=1; v<rows-1; v++){
    (void) unitnorm(&normals[idx(0,v)], subv(&ret1, &points[idx(1,v-1)], &points[idx(0,v)]),
			  subv(&ret2, &points[idx(1,v+1)], &points[idx(0,v)]));
    (void) unitnorm(&normals[idx(cols-1,v)], subv(&ret1, &points[idx(cols-2,v+1)],
							   &points[idx(cols-1,v)]),
						  subv(&ret2, &points[idx(cols-2,v-1)],
							   &points[idx(cols-1,v)]));
  }
  for (v=1; v < rows-1; v++) {
    for (u=1; u < cols-1; u++) {
      (void) unitnorm(&normals[idx(u,v)], subv(&ret1, &points[idx(u+1,v+1)],
						      &points[idx(u-1,v-1)]),
					     subv(&ret2, &points[idx(u-1,v+1)],
						      &points[idx(u+1,v-1)]));

    }
  } 
  /*
  RiBasis(RiCatmullRomBasis, RI_CATMULLROMSTEP, RiCatmullRomBasis, RI_CATMULLROMSTEP);
  */
  RiPatchMesh(RI_BILINEAR, (RtInt) cols, RI_NONPERIODIC,
                           (RtInt) rows, RI_NONPERIODIC,
	                    RI_PZ, (RtPointer) mesh,
                            RI_N, (RtPointer) normals,
                            RI_NULL);
  free(mesh);
  free(points);
  free(normals);
  return(0);
}


extern int readpgm(char *pgmfile) {
  int argc, rows, cols, ret;
  char *argv[1];
  FILE *file;
  gray **grays, maxval;

  argc = 1;
  argv[0] = pgmfile;
  pgm_init( &argc, argv );

  file = pm_openr(pgmfile);
  grays = pgm_readpgm( file, &cols, &rows, &maxval );
  pm_close( file );

  ret = parsepgm(grays, cols, rows, maxval);

  pgm_freearray(grays, rows);
  return(ret);
}







